from typing import Iterator, Optional

from ....utils import empty_iterator
from ....warehouse.abstract import WarehouseAsset
from ....warehouse.bigquery import BigQueryClient, BigQueryQueryBuilder
from .. import LookerStudioAsset
from .admin_sdk_client import USER_EMAIL_FIELD, AdminSDKClient
from .credentials import LookerStudioCredentials
from .looker_studio_api_client import LookerStudioAPIClient


class LookerStudioQueryBuilder(BigQueryQueryBuilder):
    def job_history_queries(self) -> list:
        """
        This class and method are a convenient workaround to build the
        ExtractionQueries which retrieve BigQuery's job history, but filtered on
        Looker Studio only.

        Compared to the generic BigQuery query history, only the SQL "template"
        changes. By defining this class here, this will pick the SQL file
        `queries/query.sql` located in the same directory as this file.
        """
        return super().build(WarehouseAsset.QUERY)  # type: ignore


class LookerStudioClient:
    """
    Acts as a wrapper class to fetch Looker Studio assets, which requires
    coordinating calls between the Admin SDK API and the Looker Studio API.

    If the BigQuery credentials are provided, it can also fetch the source queries
    of BigQuery data sources.
    """

    def __init__(
        self,
        credentials: LookerStudioCredentials,
        bigquery_credentials: Optional[dict] = None,
        user_emails: Optional[list[str]] = None,
    ):
        self.admin_sdk_client = AdminSDKClient(credentials)
        self.looker_studio_client = LookerStudioAPIClient(credentials)
        self.user_emails = user_emails

        self.bigquery_client: Optional[BigQueryClient] = None
        if bigquery_credentials:
            self.bigquery_client = BigQueryClient(bigquery_credentials)

    def _list_user_emails(self) -> Iterator[str]:
        """
        Lists user emails either from a provided JSON file or via the Admin SDK API.

        Using all Google Workspace users can be inefficient for large clients -
        the client might spend hours checking thousands of users for Looker Studio
        assets when only a handful actually own any. A JSON file allows
        targeting known owners instead.
        """
        if self.user_emails is not None:
            yield from self.user_emails
            return

        for user in self.admin_sdk_client.list_users():
            yield user[USER_EMAIL_FIELD]

    def _get_assets(self) -> Iterator[dict]:
        """
        Extracts reports and data sources user by user. The loop is necessary
        because the Looker Studio API can only retrieve the assets owned by a
        single user.
        """
        for user_email in self._list_user_emails():
            yield from self.looker_studio_client.fetch_user_assets(user_email)

    def _get_source_queries(self) -> Iterator[dict]:
        """
        Extracts the BigQuery jobs triggered by Looker Studio. The last job
        per data source is returned.
        """
        if not self.bigquery_client:
            return empty_iterator()

        query_builder = LookerStudioQueryBuilder(
            regions=self.bigquery_client.get_regions(),
            datasets=self.bigquery_client.get_datasets(),
            extended_regions=self.bigquery_client.get_extended_regions(),
        )

        queries = query_builder.job_history_queries()

        for query in queries:
            yield from self.bigquery_client.execute(query)

    def fetch(self, asset: LookerStudioAsset) -> Iterator[dict]:
        if asset == LookerStudioAsset.ASSETS:
            yield from self._get_assets()

        elif asset == LookerStudioAsset.SOURCE_QUERIES:
            yield from self._get_source_queries()

        elif asset == LookerStudioAsset.VIEW_ACTIVITY:
            yield from self.admin_sdk_client.list_view_events()

        else:
            raise ValueError(f"The asset {asset}, is not supported")



from pydantic import Field, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

MODE_ENV_PREFIX = "CASTOR_MODE_ANALYTICS_"


class ModeCredentials(BaseSettings):
    """Class holding Mode credentials attributes"""

    model_config = SettingsConfigDict(
        env_prefix=MODE_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    host: str
    secret: str = Field(repr=False)
    token: str
    workspace: str

    @field_validator("host", mode="before")
    @classmethod
    def _check_base_url(cls, host: str) -> str:
        return host.rstrip("/")



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import strategy  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-u", "--username", help="Strategy username")
    parser.add_argument("-p", "--password", help="Strategy password")
    parser.add_argument("-b", "--base-url", help="Strategy instance URL")
    parser.add_argument("-o", "--output", help="Directory to write to")

    parser.add_argument(
        "-i",
        "--project-ids",
        nargs="*",
        help="Optional list of project IDs",
        default=None,
    )

    strategy.extract_all(**parse_filled_arguments(parser))



from ...types import ExternalAsset


class SalesforceReportingAsset(ExternalAsset):
    """Salesforce Reporting assets"""

    DASHBOARDS = "dashboards"
    DASHBOARD_COMPONENTS = "dashboard_components"
    FOLDERS = "folders"
    REPORTS = "reports"
    REPORTS_METADATA = "reports_metadata"
    USERS = "users"



from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

THOUGHTSPOT_ENV_PREFIX = "CASTOR_THOUGHTSPOT_"


class ThoughtspotCredentials(BaseSettings):
    """Class to handle Thoughtspot rest API permissions"""

    model_config = SettingsConfigDict(
        env_prefix=THOUGHTSPOT_ENV_PREFIX,
        extra="ignore",
        populate_by_name=True,
    )

    username: str
    password: str = Field(repr=False)
    base_url: str



class ThoughtspotEndpointFactory:
    @classmethod
    def authentication(cls) -> str:
        return "api/rest/2.0/auth/token/full"

    @classmethod
    def metadata_search(cls) -> str:
        return "api/rest/2.0/metadata/search"

    @classmethod
    def liveboard_data(cls) -> str:
        return "api/rest/2.0/metadata/liveboard/data"



import logging
from collections.abc import Iterable, Iterator
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import DomoAsset
from .client import DomoClient, DomoCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: DomoClient,
) -> Iterable[tuple[DomoAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Domo"""

    logger.info("Extracting PAGES from API")
    pages = client.fetch(DomoAsset.PAGES)
    yield DomoAsset.PAGES, list(deep_serialize(pages))

    logger.info("Extracting DATASETS from API")
    datasets = list(client.fetch(DomoAsset.DATASETS))
    yield DomoAsset.DATASETS, list(deep_serialize(datasets))

    logger.info("Extracting USERS from API")
    users = client.fetch(DomoAsset.USERS)
    yield DomoAsset.USERS, list(deep_serialize(users))

    logger.info("Extracting AUDIT from API")
    audit = client.fetch(DomoAsset.AUDIT)
    yield DomoAsset.AUDIT, list(deep_serialize(audit))

    logging.info("Extracting DATAFLOWS data from API")
    dataflows = client.fetch(DomoAsset.DATAFLOWS)
    yield DomoAsset.DATAFLOWS, list(deep_serialize(dataflows))


def extract_all(**kwargs) -> None:
    """
    Extract data from Domo API
    Store the output files locally under the given output_directory
    """
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    credentials = DomoCredentials(**kwargs)
    client = DomoClient(credentials=credentials)

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts, base_url=credentials.base_url)



from uuid import UUID

from .constant import INGEST_URLS, FileType
from .enums import Zone
from .upload import _path_and_url


def test__path():
    source_id = UUID("399a8b22-3187-11ec-8d3d-0242ac130003")
    file_type = FileType.VIZ
    file_path = "filename"
    zone = Zone.EU

    path, url = _path_and_url(source_id, file_type, file_path, zone)
    assert path == f"visualization-{source_id}/{file_path}"
    assert url == f"{INGEST_URLS[Zone.EU]}/{path}"



from typing import Optional

from pydantic import BaseModel, SecretStr, field_serializer

SCOPES_NO_ACTIVITY: tuple[str, ...] = (
    "https://www.googleapis.com/auth/datastudio",
    "https://www.googleapis.com/auth/userinfo.profile",
    "https://www.googleapis.com/auth/admin.directory.user.readonly",
)

DEFAULT_SCOPES: tuple[str, ...] = (
    *SCOPES_NO_ACTIVITY,
    "https://www.googleapis.com/auth/admin.reports.audit.readonly",
)


class LookerStudioCredentials(BaseModel):
    """
    Looker Studio Credentials match the Service Account credentials JSON
    but with an additional admin_email field.
    """

    admin_email: str
    auth_provider_x509_cert_url: str
    auth_uri: str
    client_email: str
    client_id: str
    client_x509_cert_url: str
    private_key: SecretStr
    private_key_id: str
    project_id: str
    token_uri: str
    type: str

    has_view_activity_logs: Optional[bool] = True
    scopes: Optional[tuple] = DEFAULT_SCOPES

    def model_post_init(self, __context):
        """Set scopes based on has_view_activity_logs after initialization"""
        if self.has_view_activity_logs is False:
            self.scopes = SCOPES_NO_ACTIVITY

    @field_serializer("private_key")
    def dump_secret(self, pk):
        """When using model_dump, show private_key value"""
        return pk.get_secret_value()



from urllib.parse import urlsplit, urlunsplit


def url_from(
    scheme: str = "",
    netloc: str = "",
    path: str = "",
    query: str = "",
    fragment: str = "",
) -> str:
    """Constructs an url from part"""
    return urlunsplit((scheme, netloc, path, query, fragment))


def add_path(base_url: str, *paths: str) -> str:
    """Adds a path from a base_url."""

    if not is_valid(base_url):
        raise ValueError(f"Invalid base_url: {base_url}")
    base_url = _format_base_url(base_url)
    split = urlsplit(base_url)

    return url_from(
        split.scheme,
        split.netloc,
        "/".join([split.path] + [p.strip("/") for p in paths]),
        split.query,
        split.fragment,
    )


def _format_base_url(url: str) -> str:
    """Remove trailing slash in base url, if applicable."""
    if url.endswith("/"):
        return url[:-1]
    return url


def is_valid(
    url: str,
    valid_schemes: tuple[str, ...] = ("http", "https"),
) -> bool:
    """
    Simple url validation that ensures the scheme and that there is an hostname.
    Malformatted url can pass this check such as http://http://toto.com
    """
    split = urlsplit(url)
    return split.scheme in valid_schemes and bool(split.netloc)



import logging
import re
from enum import Enum
from typing import Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from sqlalchemy.engine import Connection

from ...utils import SqlalchemyClient, to_string_array, uri_encode

SNOWFLAKE_URI = "snowflake://{user}:{password}@{account}/?application=castor"
PK_URI = "snowflake://{user}:@{account}/?application=castor"
PUBLIC_ROLE = "PUBLIC"

logger = logging.getLogger(__name__)


def _scalar(connection: Connection, query: str) -> Optional[str]:
    """returns the first column of the first row, None if no row"""
    scalar = connection.execute(query).scalar()
    return None if scalar is None else str(scalar)


class UseResource(Enum):
    """resources using the USE syntax"""

    ROLE = "ROLE"
    WAREHOUSE = "WAREHOUSE"


def use(connection: Connection, resource: UseResource, name: str):
    """
    USE ROLE/WAREHOUSE do not accept SQL parameter substitution
        we do a sanity check prior to using it
    DON'T COPY THIS CODE FOR ANOTHER PURPOSE
    """
    target = resource.value
    if not re.fullmatch(r"[\w_-]+", name):
        raise NameError(f"{target} name invalid {name}")
    command = f'USE {target} "{name}"'
    logger.info(f"executing {command}")
    connection.execute(command)


def find_role(connection: Connection) -> str:
    """Finds available role within Snowflake environment"""
    current = _scalar(connection, "SELECT CURRENT_ROLE()")
    if current and current.upper() != PUBLIC_ROLE:
        logger.info(f"Role is already set: {current}")
        return current

    role = _first_non_public_role(connection)
    logger.info(f"Using first available role: {role}")
    return role


def find_warehouse(connection: Connection) -> str:
    """Finds available warehouse within Snowflake environment"""
    current = _scalar(connection, "SELECT CURRENT_WAREHOUSE()")
    if current:
        logger.info(f"Warehouse is already set: {current}")
        return current

    warehouse = _scalar(connection, "SHOW WAREHOUSES")
    if not warehouse:
        raise ConnectionError("no warehouse available")
    logger.info(f"Using first available warehouse: {warehouse}")
    return warehouse


def _first_non_public_role(connection: Connection) -> str:
    """fetch the first non public role granted to the user"""
    raw = _scalar(connection, "SELECT CURRENT_AVAILABLE_ROLES()")
    if not raw:
        raise ConnectionError("error at fetching granted roles")
    logger.info(f"available roles {raw}")
    parsed = to_string_array(raw)
    granted = {r.upper() for r in parsed} - {PUBLIC_ROLE}
    if not granted:
        raise ConnectionError("missing non public role")
    return next(iter(granted))


def _private_key(private_key: Optional[str]) -> Optional[bytes]:
    if not private_key:
        return None

    p_key = serialization.load_pem_private_key(
        bytes(private_key, "utf-8"), password=None, backend=default_backend()
    )

    pkb = p_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    return pkb


class SnowflakeClient(SqlalchemyClient):
    """snowflake client"""

    # hide information message from the connectors
    logging.getLogger("snowflake.connector").setLevel(logging.WARNING)

    def __init__(
        self,
        credentials: dict,
        warehouse: Optional[str] = None,
        role: Optional[str] = None,
    ):
        self.private_key = _private_key(credentials.get("private_key"))
        self.insecure_mode = credentials.get("insecure_mode", False)
        super().__init__(credentials)
        self._role = role
        self._warehouse = warehouse

    @staticmethod
    def name() -> str:
        return "Snowflake"

    def _engine_options(self, credentials: dict) -> dict:
        """
        Sets the engine options.
        * `insecure_mode` = True turns off OCSP checking.
            See https://community.snowflake.com/s/article/How-to-turn-off-OCSP-checking-in-Snowflake-client-drivers
        * `private_key` = Snowflake private key for key pair authentication.
            See https://docs.snowflake.com/en/user-guide/key-pair-auth
        """
        connection_args: dict = dict()

        if self.insecure_mode:
            connection_args["insecure_mode"] = True

        if self.private_key:
            connection_args["private_key"] = self.private_key

        return {"connect_args": connection_args}

    def _build_uri(self, credentials: dict) -> str:
        if self.private_key:
            return PK_URI.format(
                account=credentials["account"],
                user=credentials["user"],
            )
        return SNOWFLAKE_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            account=credentials["account"],
        )

    def connection(self) -> Connection:
        """Authenticate to snowflake instance"""
        return super().connect()

    def role(self, connection: Connection) -> None:
        """check and set role"""
        if self._role:
            logger.info(f"Role was provided in arguments: {self._role}")
            use(connection, UseResource.ROLE, self._role)
            return

        role = find_role(connection)
        use(connection, UseResource.ROLE, role)

    def warehouse(self, connection: Connection):
        """check and set warehouse"""
        if self._warehouse:
            logger.info(
                f"Warehouse was provided in arguments: {self._warehouse}",
            )
            use(connection, UseResource.WAREHOUSE, self._warehouse)
            return

        warehouse = find_warehouse(connection)
        use(connection, UseResource.WAREHOUSE, warehouse)

    def connect(self) -> Connection:
        """enhance default behaviour to check role and warehouse"""
        connection = self.connection()
        self.role(connection)
        self.warehouse(connection)
        return connection



from datetime import date

from freezegun import freeze_time

from ..abstract import TimeFilter
from .utils import (
    _day_hour_to_epoch_ms,
    build_path,
    hourly_time_filters,
    tag_label,
)


def test_build_path():
    row = {
        "bigflo": "oli",
        "laurel": "hardy",
        "dupond": "dupont",
    }
    keys = ["laurel", "dupond"]
    assert build_path(row, keys) == "hardy.dupont"


def test_tag_label():
    row = {
        "tag_name": "marketplace",
        "tag_value": "",
    }
    assert tag_label(row) == "marketplace"

    row = {
        "tag_name": "fi",
        "tag_value": "fou",
    }
    assert tag_label(row) == "fi:fou"


def test__day_hour_to_epoch_ms():
    assert _day_hour_to_epoch_ms(date(2023, 2, 14), 14) == 1676383200000


@freeze_time("2023-7-4")
def test_hourly_time_filters():
    # default is yesterday
    default_filters = [f for f in hourly_time_filters(None)]

    assert len(default_filters) == 24  # number of hours in a day

    first = default_filters[0]
    start = first["filter_by"]["query_start_time_range"]["start_time_ms"]
    last = default_filters[-1]
    end = last["filter_by"]["query_start_time_range"]["end_time_ms"]
    assert start == 1688342400000  # July 3, 2023 12:00:00 AM GMT
    assert end == 1688428800000  # July 4, 2023 12:00:00 AM GMT

    # custom time (from execution_date in DAG for example)
    time_filter = TimeFilter(day=date(2020, 10, 15))
    custom_filters = [f for f in hourly_time_filters(time_filter)]

    assert len(custom_filters) == 24

    first = custom_filters[0]
    start = first["filter_by"]["query_start_time_range"]["start_time_ms"]
    last = custom_filters[-1]
    end = last["filter_by"]["query_start_time_range"]["end_time_ms"]
    assert start == 1602720000000  # Oct 15, 2020 12:00:00 AM
    assert end == 1602806400000  # Oct 16, 2020 12:00:00 AM

    # hourly extraction: note that hour_min == hour_max
    hourly = TimeFilter(day=date(2023, 4, 14), hour_min=4, hour_max=4)
    hourly_filters = [f for f in hourly_time_filters(hourly)]
    expected_hourly = [
        {
            "filter_by": {
                "query_start_time_range": {
                    "end_time_ms": 1681448400000,  # April 14, 2023 5:00:00 AM
                    "start_time_ms": 1681444800000,  # April 14, 2023 4:00:00 AM
                }
            }
        }
    ]
    assert hourly_filters == expected_hourly



from typing import Iterator, List, TypeVar

T = TypeVar("T")


def batch_of_length(
    elements: List[T],
    batch_size: int,
) -> Iterator[List[T]]:
    """
    Split the given elements into smaller chunks
    """
    assert batch_size > 1, "batch size must be greater or equal to 1"
    element_count = len(elements)
    for index in range(0, element_count, batch_size):
        yield elements[index : min((index + batch_size), element_count)]



import csv
import os
from collections.abc import Iterator

from .column import ColumnChecker
from .enums import DataType, Issue
from .file import FileCheckerRun, FileTemplate

_TEST_FILE = "file_test_users.csv"
_TEST_FILE_VALID = "file_test_users_valid.csv"


def _content(path: str) -> Iterator[dict]:
    absolute_path = os.path.join(os.path.dirname(__file__), path)
    with open(absolute_path) as csvfile:
        yield from csv.DictReader(csvfile)


def _user_template() -> FileTemplate:
    folder_ids = {"1", "2", "3"}
    return {
        "id": ColumnChecker(is_unique=True),
        "name": ColumnChecker(),
        "gender": ColumnChecker(enum_values={"MALE", "FEMALE"}),
        "birth_date": ColumnChecker(
            data_type=DataType.DATETIME,
            is_mandatory=False,
        ),
        "description": ColumnChecker(is_mandatory=False),
        "siblings_count": ColumnChecker(
            data_type=DataType.INTEGER,
            is_mandatory=False,
        ),
        "height": ColumnChecker(data_type=DataType.FLOAT),
        "folder_id": ColumnChecker(
            data_type=DataType.INTEGER,
            foreign=folder_ids,
        ),
    }


def test__file_checker_run():
    """
    End-to-end test of FileChecker:
    - template (fake users)
    - CSV file
    - results
    """
    content = _content(_TEST_FILE)
    checker = FileCheckerRun(content, _user_template(), _TEST_FILE)

    for _ in checker.valid_rows():
        pass

    assert checker.result.valid_rows == 2
    assert checker.result.total_rows == 9
    assert checker.result.counter[Issue.MISSING_VALUE] == 1
    assert checker.result.counter[Issue.UNAUTHORIZED_VALUE] == 1
    assert checker.result.counter[Issue.WRONG_DATATYPE] == 2
    assert checker.result.counter[Issue.UNKNOWN_REFERENCE] == 1
    assert checker.result.counter[Issue.DUPLICATE_VALUE] == 1
    assert checker.result.counter[Issue.REPEATED_QUOTES] == 1

    content = _content(_TEST_FILE_VALID)
    checker = FileCheckerRun(content, _user_template(), _TEST_FILE_VALID)

    for _ in checker.valid_rows():
        pass

    assert checker.result.is_valid()



def pages_to_folder_ids(pages: list[dict]) -> set:
    """Returns all unique folder parents."""
    return {
        page["parentId"] for page in pages if page["parentType"] == "folder"
    }


def pages_to_database_ids(pages: list[dict]) -> set:
    """Returns all unique database parents."""
    return {
        page["parentId"] for page in pages if page["parentType"] == "database"
    }



from pydantic import ConfigDict, Field

from ....utils import PaginationModel

METADATA_BATCH_SIZE = 100


class ThoughtSpotPagination(PaginationModel):
    data_rows: list = Field(default_factory=list)
    record_offset: int
    record_size: int

    model_config = ConfigDict(
        populate_by_name=True,
        from_attributes=True,
    )

    def is_last(self) -> bool:
        return len(self.data_rows) < METADATA_BATCH_SIZE

    def next_page_payload(self) -> dict:
        return {"record_offset": self.record_offset + METADATA_BATCH_SIZE}

    def page_results(self) -> list:
        return self.data_rows



from typing import Optional

from ..abstract import (
    AbstractQueryBuilder,
    ExtractionQuery,
    TimeFilter,
    WarehouseAsset,
)


class RedshiftQueryBuilder(AbstractQueryBuilder):
    """
    Builds queries to extract assets from Redshift.
    """

    def __init__(
        self,
        is_serverless: bool = False,
        time_filter: Optional[TimeFilter] = None,
    ):
        super().__init__(time_filter=time_filter)
        self.is_serverless = is_serverless

    def build_query_serverless(self) -> ExtractionQuery:
        """To get the query history in Redshift Serverless, we cannot use STL tables."""
        statement = self._load_from_file("query_serverless.sql")
        params = self._time_filter.to_dict()
        return ExtractionQuery(statement, params)

    def build(self, asset: WarehouseAsset) -> list[ExtractionQuery]:
        if asset == WarehouseAsset.QUERY and self.is_serverless:
            query = self.build_query_serverless()
        else:
            query = self.build_default(asset)
        return [query]



from ...types import ExternalAsset


class SigmaAsset(ExternalAsset):
    """Sigma assets"""

    DATAMODELS = "datamodels"
    DATASETS = "datasets"
    DATASET_SOURCES = "dataset_sources"
    ELEMENTS = "elements"
    FILES = "files"
    LINEAGES = "lineages"
    MEMBERS = "members"
    QUERIES = "queries"
    WORKBOOKS = "workbooks"
    WORKBOOK_SOURCES = "workbook_sources"



import pytest

from .collection import deduplicate, filter_items, mapping_from_rows


def test__mapping_from_rows__basic_mapping():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice", 2: "Bob"}
    assert result == expected


def test__mapping_from_rows__missing_key():
    rows = [{"id": 1, "name": "Alice"}, {"name": "Bob"}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__missing_value():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2}]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__empty_list():
    rows = []
    result = mapping_from_rows(rows, "id", "name")
    expected = {}
    assert result == expected


def test__mapping_from_rows__non_existent_key_value():
    rows = [{"id": 1, "name": "Alice"}, {"id": 2, "name": "Bob"}]
    result = mapping_from_rows(rows, "nonexistent_key", "nonexistent_value")
    expected = {}
    assert result == expected


def test__mapping_from_rows__none_key_value():
    rows = [
        {"id": 1, "name": "Alice"},
        {"id": None, "name": "Bob"},
        {"id": 2, "name": None},
    ]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice"}
    assert result == expected


def test__mapping_from_rows__multiple_valid_rows():
    rows = [
        {"id": 1, "name": "Alice"},
        {"id": 2, "name": "Bob"},
        {"id": 3, "name": "Charlie"},
    ]
    result = mapping_from_rows(rows, "id", "name")
    expected = {1: "Alice", 2: "Bob", 3: "Charlie"}
    assert result == expected


def test_deduplicate():
    e1 = {"id": "1", "name": "element_1"}
    e2 = {"id": "2", "name": "element_2"}
    e3 = {"id": "3", "name": "element_3"}

    elements = [
        e1,
        e2,
        e3,
        {"id": "3", "name": "duplicate"},
        {"id": "3", "name": "duplicate"},
        {"id": "2", "name": "duplicate"},
    ]
    assert deduplicate("id", elements) == [e1, e2, e3]


def test_sqlserver_databases():
    databases = [
        "prod",
        "staging",
        "test",
    ]

    # 1. No allowed or blocked: should return all
    result1 = filter_items(databases)
    assert result1 == [
        "prod",
        "staging",
        "test",
    ]

    # 2. Block "prod": only staging and test should remain
    result2 = filter_items(databases, blocked=["prod"])
    assert result2 == ["staging", "test"]

    # 3. Only allow "staging" and "test"
    result3 = filter_items(databases, allowed=["staging", "test"])
    assert result3 == ["staging", "test"]

    # 4. allowed and blocked, should raise
    with pytest.raises(AttributeError):
        filter_items(databases, blocked=["prod"], allowed=["staging", "test"])



from .pagination import Pagination


def test_pagination():
    per_page = 20

    pagination = Pagination(per_page=per_page)

    assert pagination.number_results is None
    assert pagination.offset == 0

    pagination.increment_offset(per_page)
    assert pagination.offset == per_page
    assert pagination.needs_increment

    pagination.increment_offset(per_page)
    assert pagination.offset == per_page * 2
    assert pagination.needs_increment

    pagination.increment_offset(5)
    assert pagination.offset == per_page * 2 + 5
    assert not pagination.needs_increment



import logging
from typing import Optional, cast

import requests
from requests.auth import HTTPBasicAuth

from ....utils import retry
from ..assets import (
    ASSETS_WITH_OWNER,
    EXPORTED_FIELDS,
    ModeAnalyticsAsset as Asset,
)
from ..errors import (
    MissingPrerequisiteError,
    UnexpectedApiResponseError,
    check_errors,
)
from .constants import (
    CLIENT_NAME,
    RETRY_BASE_MS,
    RETRY_COUNT,
    RETRY_EXCEPTIONS,
    RETRY_JITTER_MS,
    RETRY_STRATEGY,
)
from .credentials import ModeCredentials

logger = logging.getLogger(__name__)

URL_TEMPLATE = "{host}/api"

RawData = list[dict]
Tokens = Optional[list[str]]


class Client:
    """
    Connect to Mode Analytics API and fetch main assets.
    https://mode.com/developer/api-reference/introduction/
    """

    def __init__(
        self,
        credentials: ModeCredentials,
    ):
        self._credentials = credentials
        self._session = requests.Session()

    def authenticate(self) -> HTTPBasicAuth:
        return HTTPBasicAuth(self._credentials.token, self._credentials.secret)

    def _check_connection(self):
        authentication = self.authenticate()
        url = self._url(with_workspace=False) + "/account"
        response = self._session.get(url, auth=authentication)
        self._handle_response(response)
        logger.info("Authentication succeeded.")

    @staticmethod
    def name() -> str:
        """return the name of the client"""
        return CLIENT_NAME

    @staticmethod
    def _handle_response(
        response: requests.Response,
        *,
        resource_name: Optional[str] = None,
    ) -> RawData:
        check_errors(response)
        result = response.json()

        if "_embedded" not in result:
            # some calls return data directly
            return result

        # most of calls return data in ["_embedded"]["resource_name"] node
        try:
            embedded = cast(dict, result["_embedded"])
            return cast(list, embedded[resource_name])
        except (ValueError, KeyError):
            raise UnexpectedApiResponseError(resource_name, result)

    def base_url(self) -> str:
        """Return base_url from credentials"""
        return f"{self._credentials.host}/{self._credentials.workspace}"

    def _url(
        self,
        with_workspace: bool = True,
        space: Optional[str] = None,
        report: Optional[str] = None,
        resource_name: Optional[str] = None,
    ) -> str:
        url = URL_TEMPLATE.format(host=self._credentials.host)
        if with_workspace:
            url += f"/{self._credentials.workspace}"
        if space:
            url += f"/spaces/{space}"
        if report:
            url += f"/reports/{report}"
        if resource_name:
            url += f"/{resource_name}"
        return url

    @retry(
        exceptions=RETRY_EXCEPTIONS,
        max_retries=RETRY_COUNT,
        base_ms=RETRY_BASE_MS,
        jitter_ms=RETRY_JITTER_MS,
        strategy=RETRY_STRATEGY,
    )
    def _call(
        self,
        *,
        with_workspace: bool = True,
        space: Optional[str] = None,
        report: Optional[str] = None,
        resource_name: Optional[str] = None,
    ) -> RawData:
        authentication = self.authenticate()
        url = self._url(with_workspace, space, report, resource_name)
        logger.info(f"Calling {url}")
        response = self._session.get(url, auth=authentication)
        return self._handle_response(response, resource_name=resource_name)

    def _reports(self, spaces: Optional[RawData]) -> RawData:
        reports: RawData = []
        # the only way to fetch reports is to loop on spaces
        # https://mode.com/developer/api-reference/analytics/reports/#listReportsInSpace
        if not spaces:
            raise MissingPrerequisiteError(
                fetched=Asset.REPORT,
                missing=Asset.COLLECTION,
            )
        for space in spaces:
            space_token = space["token"]
            # example: https://modeanalytics.com/api/{workspace}/spaces/{space_token}/reports
            result = self._call(space=space_token, resource_name="reports")
            reports.extend(result)
        return reports

    def _queries(self, reports: Optional[RawData]) -> RawData:
        queries: RawData = []
        if not reports:
            raise MissingPrerequisiteError(
                fetched=Asset.QUERY,
                missing=Asset.REPORT,
            )
        for report in reports:
            report_token = report["token"]
            result = self._call(report=report_token, resource_name="queries")
            for query in result:
                query["report_token"] = report_token
            queries.extend(result)
        return queries

    def _members(self) -> RawData:
        members: RawData = []
        # the only way to fetch members is to loop on memberships
        # https://mode.com/developer/api-reference/management/workspace-memberships/#listMemberships
        memberships = self._call(resource_name="memberships")
        for mb in memberships:
            # then we fetch users one by one, using their {username}
            # why without workspace? because users can belong to several companies
            # example: https://modeanalytics.com/api/john_doe
            result = self._call(
                resource_name=mb["member_username"],
                with_workspace=False,
            )
            members.append(cast(dict, result))
        return members

    @staticmethod
    def _post_processing(asset: Asset, data: RawData) -> RawData:
        filtered = []
        for row in data:
            if asset in ASSETS_WITH_OWNER:
                # extract creator from _links
                creator_href = row["_links"]["creator"]["href"]
                # remove "api/" to keep only the username
                row["creator"] = creator_href[5:]
            # keep only exported fields
            new = {key: row.get(key) for key in EXPORTED_FIELDS[asset]}
            filtered.append(new)
        return filtered

    def fetch(
        self,
        asset: Asset,
        *,
        additional_data: Optional[RawData] = None,
    ) -> RawData:
        """
        Fetch the given asset.

        :additional_data must be provided in certain cases
        - to fetch REPORTS, provide the list of COLLECTIONS
        - to fetch QUERIES, provide the list of REPORTS
        Otherwise MissingPrerequisiteError will be raised.

        It means that extracting all data must be executed in a certain order:
        ```
          members = client.fetch(Asset.MEMBER)
          datasources = client.fetch(Asset.DATASOURCE)
          collections = client.fetch(Asset.COLLECTION)
          reports = client.fetch(Asset.REPORT, additional_data=collections)
          queries = client.fetch(Asset.QUERY, additional_data=reports)
        ```
        """
        logger.info(f"Starting extraction for {asset.name}...")
        # specific calls
        if asset == Asset.REPORT:
            result = self._reports(spaces=additional_data)
        elif asset == Asset.MEMBER:
            result = self._members()
        elif asset == Asset.QUERY:
            result = self._queries(reports=additional_data)
        else:
            # generic calls
            # example: https://modeanalytics.com/api/{workspace}/spaces
            # example: https://modeanalytics.com/api/{workspace}/data_sources
            result = self._call(resource_name=str(asset.value))
        logger.info(f"{len(result)} rows extracted")
        return self._post_processing(asset, result)



import logging

from .....utils import SafeMode, safe_mode
from .constants import MEASURES_SESSION_PARAMS, JsonRpcMethod
from .credentials import QlikCredentials
from .error import (
    AccessDeniedError,
    AppSizeExceededError,
    PersistenceReadFailedError,
    QlikResponseKeyError,
)
from .json_rpc import JsonRpcClient
from .websocket import open_websocket

logger = logging.getLogger(__name__)


def _handle(response: dict) -> int:
    """
    Returns the object Handle from the response payload, or raises an error
    if one of the keys can't be found
    """
    try:
        return response["result"]["qReturn"]["qHandle"]
    except KeyError:
        raise QlikResponseKeyError(
            f"Could not fetch handle from response {response}"
        )


def _measure(response: dict) -> list:
    """
    Returns the measure from the response payload, or raises a custom error
    if one of the keys can't be found
    """
    try:
        return response["result"]["qLayout"]["qMeasureList"]["qItems"]
    except KeyError:
        raise QlikResponseKeyError(
            f"Could not fetch measure from response {response}"
        )


def _list_measures(client: JsonRpcClient, app_id: str) -> list:
    """
    Executes JSON-RPC messaging sequence to retrieve the list of measures
    for a given source

    references:
        - https://community.qlik.com/t5/Integration-Extension-APIs/How-to-ask-for-a-list-of-dimensions-and-measures-in-Qlik-Engine/td-p/1412156
        - https://help.qlik.com/en-US/sense-developer/May2021/Subsystems/EngineAPI/Content/Sense_EngineAPI/GenericObject/overview-generic-object.htm
    """
    response = client.send_message(
        method=JsonRpcMethod.OPEN_DOC,
        params=[app_id],
    )
    app_handle = _handle(response)

    response = client.send_message(
        method=JsonRpcMethod.CREATE_SESSION_OBJECT,
        handle=app_handle,
        params=list(MEASURES_SESSION_PARAMS),
    )
    session_handle = _handle(response)

    response = client.send_message(
        method=JsonRpcMethod.GET_LAYOUT,
        handle=session_handle,
    )
    return _measure(response)


class EngineApiClient:
    """
    Engine API client is responsible to send the sequence of messages to
    get measures using JsonRpcClient and websocket connection.
    """

    def __init__(self, credentials: QlikCredentials):
        self.credentials = credentials
        self._safe_mode = SafeMode(
            exceptions=(
                AccessDeniedError,
                AppSizeExceededError,
                PersistenceReadFailedError,
                QlikResponseKeyError,
            ),
            max_errors=float("inf"),
        )

    def measures(self, app_id: str) -> list:
        """
        Opens a websocket and pass it to a JsonRpcClient to return the list
        of measures scoped on an app_id.
        """

        @safe_mode(self._safe_mode, default=list)
        def _call(client: JsonRpcClient, app_id_: str) -> list:
            return _list_measures(client, app_id_)

        with open_websocket(
            app_id=app_id,
            server_url=self.credentials.base_url,
            api_key=self.credentials.api_key,
        ) as websocket:
            json_rpc_client = JsonRpcClient(websocket=websocket)
            return _call(json_rpc_client, app_id)



import logging
from argparse import ArgumentParser

from castor_extractor.utils import parse_filled_arguments  # type: ignore
from castor_extractor.visualization import tableau  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    user_group = parser.add_mutually_exclusive_group(required=False)
    user_group.add_argument("-u", "--user", help="Tableau user")
    user_group.add_argument("-n", "--token-name", help="Tableau token name")

    password_group = parser.add_mutually_exclusive_group(required=False)
    password_group.add_argument("-p", "--password", help="Tableau password")
    password_group.add_argument("-t", "--token", help="Tableau token")

    parser.add_argument("-b", "--server-url", help="Tableau server url")
    parser.add_argument("-i", "--site-id", help="Tableau site ID")

    parser.add_argument(
        "--skip-columns",
        dest="skip_columns",
        action="store_true",
        help="Option to avoid extracting Tableau columns, default to False",
    )

    parser.add_argument(
        "--skip-fields",
        dest="skip_fields",
        action="store_true",
        help="Option to avoid extracting Tableau fields, default to False",
    )

    parser.add_argument(
        "--with-pulse",
        dest="with_pulse",
        action="store_true",
        help="Extract Tableau Pulse assets: Metrics and Subscriptions",
    )

    parser.add_argument(
        "--page-size",
        help="Lower the pagination when request exceeds the nodes size limit",
        required=False,
    )

    parser.add_argument(
        "-ie",
        "--ignore-errors",
        action="store_true",
        dest="ignore_errors",
        help="Allow partial extraction of Fields and Columns: skip batch in case of Timeout errors",
    )
    parser.add_argument(
        "--ignore-ssl",
        action="store_true",
        dest="ignore_ssl",
        help="Disable SSL verification",
    )

    parser.add_argument("-o", "--output", help="Directory to write to")

    tableau.extract_all(**parse_filled_arguments(parser))



import logging
from typing import Optional

import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset
from ..assets import TABLEAU_PULSE_ASSETS, TableauAsset
from ..constants import CREDENTIALS_SITE_ID_KEY, DEFAULT_TIMEOUT_SECONDS
from .client_metadata_api import TableauClientMetadataApi
from .client_rest_api import TableauClientRestApi
from .client_tsc import TableauClientTSC
from .credentials import TableauCredentials

logger = logging.getLogger(__name__)

# these assets must be extracted via TableauServerClient (TSC)
_TSC_ASSETS = (
    # projects are not available in Metadata API
    TableauAsset.PROJECT,
    # view count are not available in Metadata API
    TableauAsset.USAGE,
    # only users who published content can be extracted from MetadataAPI
    TableauAsset.USER,
)

# these assets must be extracted via the REST API
_REST_API_ASSETS = (
    # Tableau Pulse assets are only available in REST API
    TableauAsset.METRIC,
    TableauAsset.METRIC_DEFINITION,
    TableauAsset.SUBSCRIPTION,
)

logging.getLogger("tableau.endpoint").setLevel(logging.WARNING)


def _add_site_id(workbooks: SerializedAsset, site_id: str) -> SerializedAsset:
    """
    Add site_id from credentials: it's necessary to compute workbook's url
    """
    for workbook in workbooks:
        workbook[CREDENTIALS_SITE_ID_KEY] = site_id
    return workbooks


def _merge_datasources(
    datasources: SerializedAsset,
    tsc_datasources: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich datasources with fields coming from TableauServerClient:
    - project_luid
    - webpage_url
    """

    mapping = {row["id"]: row for row in tsc_datasources}

    for datasource in datasources:
        if datasource["__typename"] != "PublishedDatasource":
            # embedded datasources are bound to workbooks => no project
            # embedded datasources cannot be accessed via URL => no webpage_url
            continue
        luid = datasource["luid"]
        tsc_datasource = mapping.get(luid)
        if not tsc_datasource:
            # it happens that a datasource is in Metadata API but not in TSC
            datasource["projectLuid"] = None
            datasource["webpageUrl"] = None
            continue
        datasource["projectLuid"] = tsc_datasource["project_id"]
        datasource["webpageUrl"] = tsc_datasource["webpage_url"]

    return datasources


def _merge_workbooks(
    workbooks: SerializedAsset,
    tsc_workbooks: SerializedAsset,
) -> SerializedAsset:
    """
    Enrich workbooks with fields coming from TableauServerClient:
    - project_luid
    """

    mapping = {row["id"]: row for row in tsc_workbooks}

    for workbook in workbooks:
        luid = workbook["luid"]
        tsc_workbook = mapping.get(luid)
        if not tsc_workbook:
            # it happens that a workbook is in Metadata API but not in TSC
            # in this case, we push the workbook with default project
            logger.warning(f"Workbook {luid} was not found in TSC")
            workbook["projectLuid"] = None
            continue

        workbook["projectLuid"] = tsc_workbook["project_id"]

    return workbooks


def _server(
    server_url: str,
    timeout_sec: int,
    ignore_ssl: bool = False,
) -> TSC.Server:
    verify = not ignore_ssl
    options = {"verify": verify, "timeout": timeout_sec}
    server = TSC.Server(server_url, use_server_version=True)
    server.add_http_options(options)
    return server


class TableauClient:
    """
    Connect to Tableau's API and extract assets.

    Relies on TableauServerClient (TSC) overlay for authentication
    https://tableau.github.io/server-client-python/docs/
    """

    def __init__(
        self,
        credentials: TableauCredentials,
        timeout_sec: int = DEFAULT_TIMEOUT_SECONDS,
        with_columns: bool = True,
        with_fields: bool = True,
        with_pulse: bool = False,
        override_page_size: Optional[int] = None,
        ignore_errors: bool = False,
        ignore_ssl: bool = False,
    ):
        self._credentials = credentials
        self._server = _server(
            server_url=credentials.server_url,
            timeout_sec=timeout_sec,
            ignore_ssl=ignore_ssl,
        )
        self._with_columns = with_columns
        self._with_fields = with_fields
        self._with_pulse = with_pulse
        self._ignore_errors = ignore_errors

        self._client_metadata = TableauClientMetadataApi(
            server=self._server,
            override_page_size=override_page_size,
        )
        self._client_rest = TableauClientRestApi(server=self._server)
        self._client_tsc = TableauClientTSC(server=self._server)

    @property
    def base_url(self) -> str:
        return self._credentials.server_url

    @property
    def name(self) -> str:
        return "Tableau/API"

    def _user_password_login(self) -> None:
        """Login into Tableau using user and password"""
        self._server.auth.sign_in(
            TSC.TableauAuth(
                self._credentials.user,
                self._credentials.password,
                site_id=self._credentials.site_id,
            ),
        )

    def _pat_login(self) -> None:
        """Login into Tableau using personal authentication token"""
        self._server.auth.sign_in(
            TSC.PersonalAccessTokenAuth(
                self._credentials.token_name,
                self._credentials.token,
                site_id=self._credentials.site_id,
            ),
        )

    def login(self) -> None:
        """
        Depending on the given credentials, logs-in using either:
        - user/password
        - token_name/value (Personal Access Token)
        https://help.tableau.com/current/api/rest_api/en-us/REST/rest_api_concepts_auth.htm

        Raises an error if none can be found
        """

        if self._credentials.user and self._credentials.password:
            logger.info("Logging in using user and password authentication")
            return self._user_password_login()

        if self._credentials.token_name and self._credentials.token:
            logger.info("Logging in using token authentication")
            return self._pat_login()

        raise ValueError(
            "Invalid credentials: either user/password or PAT must be provided",
        )

    def _fetch_datasources(self) -> SerializedAsset:
        asset = TableauAsset.DATASOURCE

        datasources = self._client_metadata.fetch(asset)
        tsc_datasources = self._client_tsc.fetch(asset)

        return _merge_datasources(datasources, tsc_datasources)

    def _fetch_workbooks(self) -> SerializedAsset:
        asset = TableauAsset.WORKBOOK

        site_id = self._credentials.site_id
        workbooks = self._client_metadata.fetch(asset)
        workbooks = _add_site_id(workbooks, site_id)

        workbook_projects = self._client_tsc.fetch(asset)

        return _merge_workbooks(workbooks, workbook_projects)

    def fetch(
        self,
        asset: TableauAsset,
    ) -> SerializedAsset:
        """
        Extract the given Tableau Asset
        """
        if asset in TABLEAU_PULSE_ASSETS and not self._with_pulse:
            logger.info(f"Skipping asset {asset} - Tableau Pulse de-activated")
            return []

        if asset == TableauAsset.COLUMN and not self._with_columns:
            logger.info(f"Skipping asset {asset} - deactivated columns")
            return []

        if asset == TableauAsset.FIELD and not self._with_fields:
            logger.info(f"Skipping asset {asset} - deactivated fields")
            return []

        logger.info(f"Extracting {asset.name}...")

        if asset == TableauAsset.DATASOURCE:
            # two APIs are required to extract datasources
            return self._fetch_datasources()

        if asset == TableauAsset.WORKBOOK:
            # two APIs are required to extract workbooks
            return self._fetch_workbooks()

        if asset in _TSC_ASSETS:
            # some assets can only be extracted via TSC
            return self._client_tsc.fetch(asset)

        if asset in _REST_API_ASSETS:
            # some assets can only be extracted via REST API
            return self._client_rest.fetch(asset)

        # other assets can be extracted via Metadata API
        return self._client_metadata.fetch(asset, self._ignore_errors)



import logging
from collections.abc import Iterable, Iterator
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import ConfluenceAsset
from .client import ConfluenceClient, ConfluenceCredentials
from .utils import pages_to_database_ids, pages_to_folder_ids

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: ConfluenceClient,
) -> Iterable[tuple[ConfluenceAsset, Union[list, Iterator, dict]]]:
    """Iterate over the extracted data from Confluence"""

    logger.info("Extracting USERS from API")
    users = list(deep_serialize(client.users()))
    yield ConfluenceAsset.USERS, users
    logger.info(f"Extracted {len(users)} users from API")

    logger.info("Extracting PAGES from API")
    pages = list(deep_serialize(client.pages()))
    yield ConfluenceAsset.PAGES, pages
    logger.info(f"Extracted {len(pages)} pages from API")

    folder_ids = pages_to_folder_ids(pages)
    logger.info("Extracting FOLDERS from API")
    folders = list(deep_serialize(client.folders(folder_ids)))
    yield ConfluenceAsset.FOLDERS, folders

    database_ids = pages_to_database_ids(pages)
    logger.info("Extracting DATABASES from API")
    databases = list(deep_serialize(client.databases(database_ids)))
    yield ConfluenceAsset.DATABASES, databases


def extract_all(**kwargs) -> None:
    """
    Extract data from Confluence API
    Store the output files locally under the given output_directory
    """

    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)

    credentials = ConfluenceCredentials(**kwargs)
    client = ConfluenceClient(
        credentials=credentials,
        include_archived_spaces=kwargs.get("include_archived_spaces") or False,
        include_personal_spaces=kwargs.get("include_personal_spaces") or False,
        space_ids_allowed=kwargs.get("space_ids_allowed"),
        space_ids_blocked=kwargs.get("space_ids_blocked"),
    )

    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), output_directory, ts)
        write_json(filename, data)

    es = current_timestamp()

    write_summary(output_directory, ts, duration_second=es - ts)



import logging
from argparse import ArgumentParser

from castor_extractor.warehouse import databricks  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="Databricks host")
    parser.add_argument(
        "--catalog-allowed",
        nargs="*",
        help="List of databricks catalogs that should be extracted",
    )
    parser.add_argument(
        "--catalog-blocked",
        nargs="*",
        help="List of databricks catalogs that should not be extracted",
    )
    parser.add_argument("-t", "--token", help="Databricks access token")
    parser.add_argument("-o", "--output", help="Directory to write to")
    parser.add_argument(
        "--skip-existing",
        dest="skip_existing",
        action="store_true",
        help="Skips files already extracted instead of replacing them",
    )
    parser.set_defaults(skip_existing=False)

    args = parser.parse_args()

    databricks.extract_all(
        db_allowed=args.catalog_allowed,
        db_blocked=args.catalog_blocked,
        host=args.host,
        output_directory=args.output,
        token=args.token,
        skip_existing=args.skip_existing,
    )



import logging
from collections.abc import Iterable
from typing import Union

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import PowerBiAsset
from .client import PowerbiClient, PowerbiCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: PowerbiClient,
) -> Iterable[tuple[PowerBiAsset, Union[list, dict]]]:
    for asset in PowerBiAsset:
        if asset in PowerBiAsset.optional:
            continue

        logger.info(f"Extracting {asset.name} from API")
        data = list(deep_serialize(client.fetch(asset)))
        yield asset, data
        logger.info(f"Extracted {len(data)} {asset.name} from API")


def extract_all(**kwargs) -> None:
    """
    Extract data from PowerBI REST API
    Store the output files locally under the given output_directory
    """
    _output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    creds = PowerbiCredentials(**kwargs)
    client = PowerbiClient(creds)
    ts = current_timestamp()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.name.lower(), _output_directory, ts)
        write_json(filename, data)

    write_summary(_output_directory, ts)



from .utils import pages_to_database_ids, pages_to_folder_ids


def test_pages_to_folder_ids():
    """Test the pages_to_folder_ids function."""
    pages = [
        {"id": "9", "parentId": None, "parentType": None},
        {"id": "8", "parentId": "2", "parentType": "folder"},
        {"id": "7", "parentId": "9", "parentType": "page"},
        {"id": "6", "parentId": "4", "parentType": "folder"},
        {"id": "5", "parentId": "4", "parentType": "folder"},
    ]
    expected = {"2", "4"}
    result = pages_to_folder_ids(pages)
    assert result == expected


def test_pages_to_database_id():
    """Test the pages_to_database_id function."""
    pages = [
        {"id": "1", "parentId": "db1", "parentType": "database"},
        {"id": "2", "parentId": "db2", "parentType": "database"},
        {"id": "3", "parentId": "4", "parentType": "folder"},
        {"id": "4", "parentId": None, "parentType": None},
        {"id": "5", "parentId": "db1", "parentType": "database"},
        {"id": "6", "parentId": "9", "parentType": "page"},
    ]
    expected = {"db1", "db2"}
    result = pages_to_database_ids(pages)
    assert result == expected



from typing import Optional, Union

from pydantic import AliasPath, Field

from ....utils import (
    FetchNextPageBy,
    PaginationModel,
)


class ConfluencePagination(PaginationModel):
    """Class to handle paginated results for confluence"""

    fetch_by: FetchNextPageBy = FetchNextPageBy.URL

    results: list = Field(default_factory=list)
    next_url: Optional[str] = Field(
        validation_alias=AliasPath("_links", "next"),
        default=None,
    )

    def is_last(self) -> bool:
        """Stopping condition for the pagination"""
        return self.next_url is None

    def next_page_payload(self) -> Union[str, None]:
        """Payload enabling to generate the request for the next page"""
        return self.next_url

    def page_results(self) -> list:
        """List of results of the current page"""
        return self.results



import logging
from collections.abc import Iterable

from ...utils import (
    OUTPUT_DIR,
    current_timestamp,
    deep_serialize,
    from_env,
    get_output_filename,
    write_json,
    write_summary,
)
from .assets import TableauAsset
from .client import TableauClient, TableauCredentials

logger = logging.getLogger(__name__)


def iterate_all_data(
    client: TableauClient,
) -> Iterable[tuple[TableauAsset, list]]:
    """Iterate over the extracted Data from Tableau"""

    for asset in TableauAsset:
        data = client.fetch(asset)
        yield asset, deep_serialize(data)


def extract_all(**kwargs) -> None:
    """
    Extract Data From tableau and store it locally in files under the
    output_directory
    """
    output_directory = kwargs.get("output") or from_env(OUTPUT_DIR)
    with_columns = not kwargs.get("skip_columns")
    with_fields = not kwargs.get("skip_fields")
    with_pulse = kwargs.get("with_pulse") or False
    page_size = kwargs.get("page_size")
    ignore_errors = kwargs.get("ignore_errors") or False
    ignore_ssl = kwargs.get("ignore_ssl") or False
    timestamp = current_timestamp()

    credentials = TableauCredentials(**kwargs)
    client = TableauClient(
        credentials,
        with_columns=with_columns,
        with_fields=with_fields,
        with_pulse=with_pulse,
        override_page_size=page_size,
        ignore_errors=ignore_errors,
        ignore_ssl=ignore_ssl,
    )
    client.login()

    for key, data in iterate_all_data(client):
        filename = get_output_filename(key.value, output_directory, timestamp)
        write_json(filename, data)

    write_summary(
        output_directory,
        timestamp,
        base_url=client.base_url,
        client_name=client.name,
    )
